unit USystray;

interface
uses
  Windows, Messages, Classes, Graphics, Controls, ShellAPI, Menus, ExtCtrls,
  Forms;

const
  sWM_TOOLTRAYICON: PChar = 'WM_TOOLTRAYICON';
  sWM_RESETTOOLTIP: PChar = 'WM_RESETTOOLTIP';
  sWM_TASKBARCREATED: PChar = 'TaskbarCreated';
  iDEFAULT_ANIMATE_INTERVAL = 100;

var
  WM_TOOLTRAYICON: UINT = 0;
  WM_RESETTOOLTIP: UINT = 0;
  WM_TASKBARCREATED: Cardinal;

type
  TModIconTypes = (modUnknown, modMessage, modIcon, modToolTip);
  TModIconOps = set of TModIconTypes;

  TTrayIcon = class(TComponent)
  private
    //Private declarations
    IconData: TNotifyIconData;
    FActive,
      FShowDesigning: Boolean;
    FIcon,
      FOriginalIcon: TIcon;
    FToolTip: string;
    FWindowHandle: HWND;
    FTimer: TTimer;
    FOnClick,
      FOnDblClick: TNotifyEvent;
    FOnRightClick: TMouseEvent;
    FPopupMenu: TPopupMenu;
    function AddIcon: Boolean;
    function ModifyIcon(modIconOps: TModIconOps): Boolean;
    function DeleteIcon: Boolean;
    procedure SetActive(bValue: Boolean);
    procedure SetShowDesigning(bValue: Boolean);
    procedure SetIcon(icnValue: TIcon);
    procedure SetToolTip(const sValue: string);
    procedure WndProc(var msg: TMessage);
    procedure FillDataStructure(modIconOps: TModIconOps);
    procedure TimerTimer(Sender: TObject);
    procedure DefaultEvent(Sender: TObject);
  protected
    procedure DoDblClick(Sender: TObject);
    procedure DoClick(Sender: TObject);
    procedure DoRightClick(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ChangeIcon(NewIcon: TIcon; TimerInterval: Integer; NextIcon: TIcon);
    property Handle: HWND read FWindowHandle;
  published
    property Active: Boolean read FActive write SetActive;
    property ShowDesigning: Boolean read FShowDesigning write SetShowDesigning;
    property Icon: TIcon read FIcon write SetIcon;
    property ToolTip: string read FTooltip write SetToolTip;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnRightClick: TMouseEvent read FOnRightClick write FOnRightClick;
    property PopupMenu: TPopupMenu read FPopupMenu write FPopupMenu;
  end;

implementation
uses
  Dialogs, SysUtils;

const
  cCR = #13;

{$R USystray.res}

  //------------------------------------------------------------------------------

constructor TTrayIcon.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FWindowHandle := Classes.AllocateHWnd(WndProc);
  WM_TOOLTRAYICON := RegisterWindowMessage(sWM_TOOLTRAYICON);
  WM_RESETTOOLTIP := RegisterWindowMessage(sWM_RESETTOOLTIP);
  WM_TASKBARCREATED := RegisterWindowMessage(sWM_TASKBARCREATED);
  FIcon := TIcon.Create;
  FOriginalIcon := TIcon.Create;
  FTimer := TTimer.Create(Self);
  with FTimer do
  begin
    Enabled := False;
    OnTimer := TimerTimer;
    Interval := iDEFAULT_ANIMATE_INTERVAL;
  end;
  if not (csDesigning in ComponentState) then
  begin
    FOnDblClick := DefaultEvent; // Default behaviour
  end;
end;

//------------------------------------------------------------------------------

destructor TTrayIcon.Destroy;
begin
  if (not (csDesigning in ComponentState) and FActive) or
    ((csDesigning in ComponentState) and FShowDesigning) then
    DeleteIcon;
  FTimer.Free;
  FOriginalIcon.Free;
  FIcon.Free;
  Classes.DeAllocateHWnd(FWindowHandle);
  inherited Destroy;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.DoDblClick(Sender: TObject);
begin
  if Assigned(FOnDblClick) then
    FOnDblClick(Sender);
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.DoClick(Sender: TObject);
begin
  if Assigned(FOnClick) then
    FOnClick(Sender);
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.DoRightClick(Sender: TObject);
var
  MouseCo: TPoint;
begin
  GetCursorPos(MouseCo);
  if Assigned(FPopupMenu) then
  begin
    SetForegroundWindow(Application.Handle);
    Application.ProcessMessages;
    FPopupMenu.Popup(MouseCo.x, MouseCo.y);
  end;
  if Assigned(FOnRightClick) then
    FOnRightClick(Sender, mbRight, [], MouseCo.x, MouseCo.y);
end;

//------------------------------------------------------------------------------

function TTrayIcon.AddIcon: Boolean;
begin
  // (use application icon if not assigned)
  if FIcon.Empty then
  begin
    FIcon.Assign(Application.Icon);
    ModifyIcon([modIcon]);
  end;
  FillDataStructure([modMessage, modIcon, modToolTip]);
  Result := Shell_NotifyIcon(NIM_ADD, @IconData);
  //For some reason, if there is no tool tip set up, then the icon
  //doesn't display.  This fixes that.
  if FToolTip = '' then
    PostMessage(FWindowHandle, WM_RESETTOOLTIP, 0, 0);
end;

//------------------------------------------------------------------------------

function TTrayIcon.ModifyIcon(modIconOps: TModIconOps): Boolean;
begin
  FillDataStructure(modIconOps);
  if FActive then
    Result := Shell_NotifyIcon(NIM_MODIFY, @IconData)
  else
    Result := True;
end;

//------------------------------------------------------------------------------

function TTrayIcon.DeleteIcon: Boolean;
begin
  Result := Shell_NotifyIcon(NIM_DELETE, @IconData);
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.SetActive(bValue: Boolean);
begin
  if bValue <> FActive then
  begin
    FActive := bValue;
    if not (csdesigning in ComponentState) then
      if FActive then
        AddIcon
      else
        DeleteIcon;
  end;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.SetShowDesigning(bValue: Boolean);
begin
  if csDesigning in ComponentState then
    if bValue <> FShowDesigning then
    begin
      FShowDesigning := bValue;
      if FShowDesigning then
        AddIcon
      else
        DeleteIcon;
    end;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.SetIcon(icnValue: TIcon);
begin
  if icnValue <> FIcon then
  begin
    FIcon.Assign(icnValue);
    ModifyIcon([modIcon]);
  end;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.SetToolTip(const sValue: string);
begin
  FToolTip := Copy(sValue, 1, 62);
  ModifyIcon([modToolTip, modIcon]);
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.WndProc(var msg: TMessage);
begin
  with msg do
    if msg = WM_RESETTOOLTIP then
      SetToolTip(FToolTip)
    else if msg = WM_TOOLTRAYICON then
      case lParam of
        WM_LBUTTONDBLCLK: DoDblClick(Self);
        WM_LBUTTONUP: DoClick(Self);
        WM_RBUTTONUP: DoRightClick(Self);
      end
    // Show the tray icon if the taskbar has been re-created after an Explorer crash
    else if (msg = WM_TASKBARCREATED) and FActive then
    begin
      DeleteIcon;
      AddIcon;
    end
    else //Handle all messages with the default handler
      Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.FillDataStructure(modIconOps: TModIconOps);
begin
  with IconData do
  begin
    cbSize := SizeOf(TNotifyIconData);
    wnd := FWindowHandle;
    uID := 0; //is not passed in with message so make it 0
    uFlags := 0;
    if modMessage in modIconOps then
    begin
      uFlags := uFlags or NIF_MESSAGE;
      uCallbackMessage := WM_TOOLTRAYICON;
    end;
    if modIcon in modIconOps then
    begin
      uFlags := uFlags or NIF_ICON;
      hIcon := FIcon.Handle;
    end;
    if modToolTip in modIconOps then
    begin
      uFlags := uFlags or NIF_TIP;
      StrPCopy(szTip, FToolTip);
    end;
  end;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.TimerTimer(Sender: TObject);
begin
  SetIcon(FOriginalIcon);
  FTimer.Enabled := False;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.DefaultEvent(Sender: TObject);
begin
  //do nothing;
end;

//------------------------------------------------------------------------------

procedure TTrayIcon.ChangeIcon(NewIcon: TIcon; TimerInterval: Integer; NextIcon: TIcon);
begin
  if Assigned(NextIcon) and (NextIcon <> FOriginalIcon) then
    FOriginalIcon.Assign(NextIcon);
  if Assigned(NewIcon) and (NewIcon <> FIcon) then
    SetIcon(NewIcon);
  if TimerInterval > 0 then
    FTimer.Interval := TimerInterval;
  FTimer.Enabled := True;
end;

//------------------------------------------------------------------------------
end.

